; ATtiny15 MBUS interface for NOKIA phone
; Simplified for debugging :)

.include "tn15def.inc"

; PINS
.equ	MBUS	= 0	; Must be pulled-up externally

.def	TEMP	= r16
.def	BYTE	= r17
.def	CHSUM	= r18
.def	DELAY	= r19
.def	PACKNUM	= r20
.def	DATA	= r21
.def	LEN_L	= r24
.def	LEN_H	= r25

.org 0x1ff
osc_const: .db 0x00 ; place for individual chip constant (convenient to use with PonyProg)

.org	0
	rjmp	RESET
	rjmp	INTERRUPT0
	rjmp	PIN_CHANGE
	rjmp	T1_COMPARE
	rjmp	T1_OVERFLOW
	rjmp	T0_OVERFLOW
	rjmp	EE_READY
	rjmp	COMPARATOR
	rjmp	ADC_COMPLETE
INTERRUPT0:
PIN_CHANGE:
T1_COMPARE:
T1_OVERFLOW:
T0_OVERFLOW:
EE_READY:
COMPARATOR:
ADC_COMPLETE:
	reti

;===================================================
;
; Starting here
; 
RESET:

; IMPORTANT!!!
; Place here a constant hardcoded in YOUR chip
;	ldi	TEMP,138
;	out	OSCCAL,TEMP	; Adjust internal oscillator !!!

; advanced way: write predefined osc constant to program memory without recompiling the program.
	ldi ZH, high(osc_const*2)
	ldi ZL, low(osc_const*2)
	lpm
	out OSCCAL, r0
	
; IMPORTANT!!!
; MBUS pin controls by DDRB.MBUS to work as an open-collector output
; PORTB.MBUS must be always zero to enable open-collector emulation
; Note that DDRB-bit is inverted (DDRB=0 -> pin is pulled up)
	cbi	PORTB,MBUS	; Turn pull-up off !!!
	cbi	DDRB,MBUS	; Must have external pullup !!!
	
;============================================================
;  Pins PORTB 1-4 are used as outputs for DEBUG
	sbi	DDRB,1		; Output pins for DEBUG
	sbi	DDRB,2
	sbi	DDRB,3
	sbi	DDRB,4
;============================================================



MAIN_LOOP:
	rcall	MB_RCV_DLR_PACKET	; Wait for DLR packet

;============================================================
;  Output DATA 0-3 to pins PORTB 1-4 for DEBUG
;---------------------------- bit 0
	cbi	PORTB,1
	sbrc	DATA,0
	sbi	PORTB,1
;---------------------------- bit 1
	cbi	PORTB,2
	sbrc	DATA,1
	sbi	PORTB,2
;---------------------------- bit 2
	cbi	PORTB,3
	sbrc	DATA,2
	sbi	PORTB,3
;---------------------------- bit 3
	cbi	PORTB,4
	sbrc	DATA,3
	sbi	PORTB,4
;============================================================
	rcall	DLY_bit			; Delay (Don't know why :)
	rcall	DLY_bit
	rcall	DLY_bit
	rcall	DLY_bit
	rcall	DLY_bit
	rcall	DLY_bit
	rcall	DLY_bit
	rcall	DLY_bit
	rcall	DLY_bit
	rcall	DLY_bit

	rcall	DLY_bit
	rcall	DLY_bit
	rcall	DLY_bit
	rcall	DLY_bit
	rcall	DLY_bit
	rcall	DLY_bit
	rcall	DLY_bit
	rcall	DLY_bit
	rcall	DLY_bit
	rcall	DLY_bit

	rcall	DLY_bit
	rcall	DLY_bit
	rcall	DLY_bit
	rcall	DLY_bit
	rcall	DLY_bit
	rcall	DLY_bit
	rcall	DLY_bit
	rcall	DLY_bit
	rcall	DLY_bit
	rcall	DLY_bit

	rcall	DLY_bit
	rcall	DLY_bit
	rcall	DLY_bit
	rcall	DLY_bit
	rcall	DLY_bit
	rcall	DLY_bit
	rcall	DLY_bit
	rcall	DLY_bit
	rcall	DLY_bit
	rcall	DLY_bit

	rcall	MB_SND_DLR_PACKET	; Send DLR ack
	rjmp	MAIN_LOOP

;===================================================
;
; Wait for a valid DLR packet from MBUS
; Does not detect any error !!!
; Returns: PACKNUM = packet number
;          DATA    = last data byte
;          LEN_H   = 0
;          LEN_L   = 0
;
MB_RCV_DLR_PACKET:
	clr	CHSUM
	rcall	MB_GET
	cpi	BYTE,0x1f	; Wait for MBUS packet id
	brne	MB_RCV_DLR_PACKET
	rcall	MB_GET		; Wait for Destination = DLR
	cpi	BYTE,0x48
	brne	MB_RCV_DLR_PACKET
	rcall	MB_GET		; Wait for Source = PHONE
	cpi	BYTE,0x00
	brne	MB_RCV_DLR_PACKET
	rcall	MB_GET
	cpi	BYTE,0x4E	; Command = Modem Status ???
	brne	MB_RCV_DLR_PACKET

	rcall	MB_GET		; Data
	rcall	MB_GET		; Data
	rcall	MB_GET		; Data
	rcall	MB_GET		; Data

	mov	DATA,BYTE
;============
	rcall	MB_GET		; Packet Number
	mov	PACKNUM,BYTE
	rcall	MB_GET		; XOR checksum
	
; Checksum must be equal to zero here - just ignored
	ret
;===================================================

;===================================================
;
; Sends a DLR3 packet to MBUS
; Argument: PACKNUM = packet number
;
MB_SND_DLR_PACKET:
	clr	CHSUM
	ldi	BYTE,0x1F	; MBUS packet id
	rcall	MB_PUT
	rcall	DLY_bit
	rcall	DLY_half
	ldi	BYTE,0x00	; Destination = PHONE
	rcall	MB_PUT
	rcall	DLY_bit
	rcall	DLY_half
	ldi	BYTE,0x48	; Source = DLR
	rcall	MB_PUT
	rcall	DLY_bit
	rcall	DLY_half
	ldi	BYTE,0x7F	; Command = Modem Ack ???
	rcall	MB_PUT
	rcall	DLY_bit
	rcall	DLY_half
	mov	BYTE,PACKNUM	; Packet number
	rcall	MB_PUT
	rcall	DLY_bit
	rcall	DLY_half
	mov	BYTE,CHSUM	; XOR checksum
	rcall	MB_PUT
	ret
;===================================================
	
;===================================================
;
; Sends a single byte to MBUS
; Argument: BYTE = byte to send
; Returns:  TEMP changed
; Returns:  CHSUM updated
;
MB_PUT:
	eor	CHSUM,BYTE	; update checksum
	clr	TEMP		; Ones count for parity
	
	sbi	DDRB,MBUS	; Send START bit = 0
	nop
	rcall	DLY_snd		; Wait bit period
	sbrc	BYTE,0		; Check bit
	rjmp	set0		;
	sbi	DDRB,MBUS	; Send bit = 0
	rjmp	wait0		;
set0:	cbi	DDRB,MBUS	; Send bit = 1
	inc	TEMP		;
wait0:	rcall	DLY_snd		; Wait bit period
	sbrc	BYTE,1		; Check bit
	rjmp	set1		;
	sbi	DDRB,MBUS	; Send bit = 0
	rjmp	wait1		;
set1:	cbi	DDRB,MBUS	; Send bit = 1
	inc	TEMP		;
wait1:	rcall	DLY_snd		; Wait bit period
	sbrc	BYTE,2		; Check bit
	rjmp	set2		;
	sbi	DDRB,MBUS	; Send bit = 0
	rjmp	wait2		;
set2:	cbi	DDRB,MBUS	; Send bit = 1
	inc	TEMP		;
wait2:	rcall	DLY_snd		; Wait bit period
	sbrc	BYTE,3		; Check bit
	rjmp	set3		;
	sbi	DDRB,MBUS	; Send bit = 0
	rjmp	wait3		;
set3:	cbi	DDRB,MBUS	; Send bit = 1
	inc	TEMP		;
wait3:	rcall	DLY_snd		; Wait bit period
	sbrc	BYTE,4		; Check bit
	rjmp	set4		;
	sbi	DDRB,MBUS	; Send bit = 0
	rjmp	wait4		;
set4:	cbi	DDRB,MBUS	; Send bit = 1
	inc	TEMP		;
wait4:	rcall	DLY_snd		; Wait bit period
	sbrc	BYTE,5		; Check bit
	rjmp	set5		;
	sbi	DDRB,MBUS	; Send bit = 0
	rjmp	wait5		;
set5:	cbi	DDRB,MBUS	; Send bit = 1
	inc	TEMP		;
wait5:	rcall	DLY_snd		; Wait bit period
	sbrc	BYTE,6		; Check bit
	rjmp	set6		;
	sbi	DDRB,MBUS	; Send bit = 0
	rjmp	wait6		;
set6:	cbi	DDRB,MBUS	; Send bit = 1
	inc	TEMP		;
wait6:	rcall	DLY_snd		; Wait bit period
	sbrc	BYTE,7		; Check bit
	rjmp	set7		;
	sbi	DDRB,MBUS	; Send bit = 0
	rjmp	wait7		;
set7:	cbi	DDRB,MBUS	; Send bit = 1
	inc	TEMP		;
wait7:	rcall	DLY_snd		; Wait bit period
	sbrs	TEMP,0		; Check PARITY
	rjmp	set8		;
	sbi	DDRB,MBUS	; Send bit = 0
	rjmp	wait8		;
set8:	cbi	DDRB,MBUS	; Send bit = 1
	inc	TEMP		;
wait8:	rcall	DLY_snd		; Wait bit period
	nop
	nop
	nop
	cbi	DDRB,MBUS	; Send STOP bit = 1
	rcall	DLY_snd		; Wait bit period
	ret
;===================================================

;===================================================
;
; Waits & receives a single byte from MBUS
; Without any error detection !!!
; Returns BYTE = received byte
;
MB_GET:
	clr	BYTE
	sbic	PINB,MBUS	; Wait for START bit
	rjmp	MB_GET
	rcall	DLY_half	; Wait half-bit period
	sbic	PINB,MBUS	; Check START bit
	rjmp	MB_GET
	rcall	DLY_bit		; Wait bit period
	sbic	PINB,MBUS	; Get bit 0
	sbr	BYTE,1
	rcall	DLY_bit		; Wait bit period
	sbic	PINB,MBUS	; Get bit 1
	sbr	BYTE,2
	rcall	DLY_bit		; Wait bit period
	sbic	PINB,MBUS	; Get bit 2
	sbr	BYTE,4
	rcall	DLY_bit		; Wait bit period
	sbic	PINB,MBUS	; Get bit 3
	sbr	BYTE,8
	rcall	DLY_bit		; Wait bit period
	sbic	PINB,MBUS	; Get bit 4
	sbr	BYTE,16
	rcall	DLY_bit		; Wait bit period
	sbic	PINB,MBUS	; Get bit 5
	sbr	BYTE,32
	rcall	DLY_bit		; Wait bit period
	sbic	PINB,MBUS	; Get bit 6
	sbr	BYTE,64
	rcall	DLY_bit		; Wait bit period
	sbic	PINB,MBUS	; Get bit 7
	sbr	BYTE,128
	rcall	DLY_bit		; Wait bit period
				; Parity bit - ignored
get:	sbis	PINB,MBUS	; Wait for STOP bit
	rjmp	get
	eor	CHSUM,BYTE	; update checksum
	ret
;===================================================
	
;===================================================
;
; DELAY routines for 9600 @ 1,6 MHz
; 
; 
DLY_snd:			; Calculated - 50 = 158 cycles
	ldi	DELAY,48
	nop
	nop
	rjmp	dly
DLY_bit:			; Calculated - 52 = 162 cycles
	ldi	DELAY,50
	rjmp	dly
DLY_half:
	ldi	DELAY,25
dly:	dec	DELAY
	brne	dly
	ret
;===================================================
	